import { __classPrivateFieldGet, __classPrivateFieldSet } from 'tslib';
import { isDefined, assert, convertTime, Deferred, mapAsKeys, wrapMethod } from '@s-libs/js-core';
import { isString, isRegExp, remove, isUndefined, bindKey, clone, forOwn, nth, isEqual, functions, keys } from '@s-libs/micro-dash';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import * as i0 from '@angular/core';
import { ErrorHandler, Injectable, ApplicationRef, Component, ApplicationInitStatus } from '@angular/core';
import { flush, TestBed, flushMicrotasks, tick, discardPeriodicTasks, fakeAsync } from '@angular/core/testing';
import { HarnessEnvironment, ContentContainerComponentHarness } from '@angular/cdk/testing';
import { UnitTestElement } from '@angular/cdk/testing/testbed';
import { By } from '@angular/platform-browser';
import { NoopAnimationsModule } from '@angular/platform-browser/animations';
import { TestScheduler } from 'rxjs/testing';

function buildErrorMessage({ matchType, itemType, stringifiedUserInput, matches, }) {
    let message = `Expected ${matchType} ${itemType}(s)`;
    if (isDefined(stringifiedUserInput)) {
        message += ` for criterion "${stringifiedUserInput}"`;
    }
    message += `, found ${matches.length}`;
    return message;
}

var _MockErrorHandler_errors;
/**
 * An error handler to be used in tests, that keeps track of all errors it handles and allows for expectations to run on them.
 *
 * ```ts
 * it('tracks errors', () => {
 *   const ctx = new AngularContext();
 *   ctx.run(() => {
 *     // test something that is supposed to throw an error
 *     ctx.inject(ErrorHandler).handleError(new Error('special message'));
 *
 *     // expect that it did
 *     ctx.inject(MockErrorHandler).expectOne('special message');
 *   });
 * });
 * ```
 */
class MockErrorHandler extends ErrorHandler {
    constructor() {
        super(...arguments);
        _MockErrorHandler_errors.set(this, []);
    }
    /**
     * Convenience method to put in a `provide` array, to override Angular's default error handler. You do not need to use this if you are using {@linkcode AngularContext}, which automatically provides it.
     *
     * ```ts
     * TestBed.configureTestingModule({
     *   providers: [MockErrorHandler.overrideProvider()],
     * });
     * ```
     */
    static overrideProvider() {
        return { provide: ErrorHandler, useExisting: MockErrorHandler };
    }
    /**
     * In addition to tracking the error, this call's Angular's [ErrorHandler.handleError]{@linkcode https://angular.io/api/core/ErrorHandler#handleError}, prints the error to the console and may print additional information that could be helpful for finding the source of the error.
     */
    handleError(error) {
        super.handleError(error);
        __classPrivateFieldGet(this, _MockErrorHandler_errors, "f").push(error);
    }
    /**
     * Expect that a single error was handled that matches the given message or predicate, and return it. If no such error was handled, or more than one, fail with a message including `description`, if provided.
     */
    expectOne(match, description) {
        const matches = this.match(match);
        if (matches.length !== 1) {
            throw new Error(buildErrorMessage({
                matchType: 'one matching',
                itemType: 'error',
                matches,
                stringifiedUserInput: stringifyUserInput(match, description),
            }));
        }
        return matches[0];
    }
    /**
     * Expect that no errors were handled which match the given message or predicate. If a matching error was handled, fail with a message including `description`, if provided.
     */
    expectNone(match, description) {
        const matches = this.match(match);
        if (matches.length > 0) {
            throw new Error(buildErrorMessage({
                matchType: 'zero matching',
                itemType: 'error',
                matches,
                stringifiedUserInput: stringifyUserInput(match, description),
            }));
        }
    }
    /**
     * Search for errors that match the given message or predicate, without any expectations.
     */
    match(match) {
        if (isString(match)) {
            const message = match;
            match = (error) => error?.message === message;
        }
        else if (isRegExp(match)) {
            const regex = match;
            match = (error) => regex.test(error?.message);
        }
        return remove(__classPrivateFieldGet(this, _MockErrorHandler_errors, "f"), match);
    }
    /**
     * Verify that no unmatched errors are outstanding. If any are, fail with a message indicating which calls were not matched.
     */
    verify() {
        const count = __classPrivateFieldGet(this, _MockErrorHandler_errors, "f").length;
        if (count > 0) {
            throw new Error(`Expected no error(s), found ${count}`);
        }
    }
}
_MockErrorHandler_errors = new WeakMap();
MockErrorHandler.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.0.3", ngImport: i0, type: MockErrorHandler, deps: null, target: i0.ɵɵFactoryTarget.Injectable });
MockErrorHandler.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "14.0.3", ngImport: i0, type: MockErrorHandler, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.0.3", ngImport: i0, type: MockErrorHandler, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }] });
function stringifyUserInput(match, description) {
    if (isUndefined(description)) {
        if (isString(match)) {
            description = 'Match by string: ' + match;
        }
        else if (isRegExp(match)) {
            description = 'Match by regexp: ' + match;
        }
        else {
            description = 'Match by function: ' + match.name;
        }
    }
    return description;
}

class FakeAsyncHarnessEnvironment extends HarnessEnvironment {
    constructor(rawRootElement, ctx) {
        super(rawRootElement);
        this.ctx = ctx;
    }
    static documentRootLoader(ctx) {
        return new FakeAsyncHarnessEnvironment(document.body, ctx);
    }
    async waitForTasksOutsideAngular() {
        flush();
    }
    async forceStabilize() {
        this.ctx.tick();
    }
    createEnvironment(element) {
        return new FakeAsyncHarnessEnvironment(element, this.ctx);
    }
    createTestElement(element) {
        return new UnitTestElement(element, bindKey(this, 'forceStabilize'));
    }
    async getAllRawElements(selector) {
        return Array.from(this.rawRootElement.querySelectorAll(selector));
    }
    getDocumentRoot() {
        return document.body;
    }
}

var _AngularContext_instances, _a, _AngularContext_current, _AngularContext_loader, _AngularContext_runWithMockedTime;
function extendMetadata(metadata, toAdd) {
    const result = clone(metadata);
    forOwn(toAdd, (val, key) => {
        const existing = result[key];
        if (isUndefined(existing)) {
            result[key] = val;
        }
        else if (key === 'imports') {
            // to allow ComponentContext to unconditionally disable animations, added imports override previous imports
            result[key] = [result[key], val];
        }
        else {
            // but for most things we want to let what comes in from subclasses and users win
            result[key] = [val, result[key]];
        }
    });
    return result;
}
/**
 * Provides the foundation for an opinionated testing pattern.
 * - All test are run in the `fakeAsync` zone. This gives you full control over
 *   the timing of everything by default.
 * - Variables that are initialized for each test exist in a context that is
 *   thrown away, so they cannot leak between tests.
 * - Clearly separates initialization code from the test itself.
 * - Gives control over the simulated date & time with a single line of code.
 * - Automatically includes [HttpClientTestingModule]{@link https://angular.io/api/common/http/testing/HttpClientTestingModule} to stub network requests without additional setup.
 * - Always verifies that no unexpected http requests were made.
 * - Automatically discards periodic tasks and flushes pending timers at the end of each test to avoid the error "X timer(s) still in the queue".
 *
 * This example tests a simple service that uses `HttpClient`, and is tested by using `AngularContext` directly. More often `AngularContext` will be used as a super class. See {@link ComponentContext} for more common use cases.
 *
 * ```ts
 * // This is the class we will test.
 * @Injectable({ providedIn: 'root' })
 * class MemoriesService {
 *   constructor(private httpClient: HttpClient) {}
 *
 *   getLastYearToday(): Observable<any> {
 *     const datetime = new Date();
 *     datetime.setFullYear(datetime.getFullYear() - 1);
 *     const date = datetime.toISOString().split('T')[0];
 *     return this.httpClient.get(`http://example.com/post-from/${date}`);
 *   }
 * }
 *
 * describe('MemoriesService', () => {
 *   // Tests should have exactly 1 variable outside an "it": `ctx`.
 *   let ctx: AngularContext;
 *   beforeEach(() => {
 *     ctx = new AngularContext();
 *   });
 *
 *   it('requests a post from 1 year ago', () => {
 *     // Before calling `run`, set up any context variables this test needs.
 *     ctx.startTime = new Date('2004-02-16T10:15:00.000Z');
 *
 *     // Pass the test itself as a callback to `run()`.
 *     ctx.run(() => {
 *       const httpBackend = ctx.inject(HttpTestingController);
 *       const myService = ctx.inject(MemoriesService);
 *
 *       myService.getLastYearToday().subscribe();
 *
 *       httpBackend.expectOne('http://example.com/post-from/2003-02-16');
 *     });
 *   });
 * });
 * ```
 */
class AngularContext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link HttpClientTestingModule} for you.
     */
    constructor(moduleMetadata = {}) {
        _AngularContext_instances.add(this);
        /**
         * Set this before calling `run()` to mock the time at which the test starts.
         */
        this.startTime = new Date();
        _AngularContext_loader.set(this, FakeAsyncHarnessEnvironment.documentRootLoader(this));
        assert(!__classPrivateFieldGet(AngularContext, _a, "f", _AngularContext_current), 'There is already another AngularContext in use (or it was not cleaned up)');
        __classPrivateFieldSet(AngularContext, _a, this, "f", _AngularContext_current);
        TestBed.configureTestingModule(extendMetadata(moduleMetadata, {
            imports: [HttpClientTestingModule],
            providers: [MockErrorHandler.overrideProvider()],
        }));
    }
    /**
     * Returns the current `AngularContext` that is in use, or `undefined` if there is not one. A context is defined to be "in use" from the time it is constructed until after its `run()` method completes.
     */
    static getCurrent() {
        return __classPrivateFieldGet(AngularContext, _a, "f", _AngularContext_current);
    }
    /**
     * Runs `test` in a `fakeAsync` zone. It can use async/await, but be sure anything you `await` is already due to execute (e.g. if a timeout is due in 3 seconds, call `.tick(3000)` before `await`ing its result).
     *
     * Also runs the following in this order, all within the same zone:
     *
     * 1. `this.init()`
     * 2. `test()`
     * 3. `this.verifyPostTestConditions()`
     * 4. `this.cleanUp()`
     */
    run(test) {
        __classPrivateFieldGet(this, _AngularContext_instances, "m", _AngularContext_runWithMockedTime).call(this, () => {
            this.init();
            try {
                test();
                this.tick();
                this.verifyPostTestConditions();
            }
            finally {
                this.cleanUp();
            }
        });
    }
    /**
     * Gets a service or other injectable from the root injector. This implementation is a simple pass-through to [TestBed.inject()]{@linkcode https://angular.io/api/core/testing/TestBed#inject}, but subclasses may provide their own implementation. It is recommended to use this in your tests instead of using `TestBed` directly.
     */
    inject(token) {
        return TestBed.inject(token);
    }
    /**
     * A convenience method to determine whether a given harness can be found.
     */
    async hasHarness(query) {
        const found = await this.getAllHarnesses(query);
        return found.length > 0;
    }
    /**
     * Gets a component harness, wrapped for use in a fakeAsync test so that you do not need to `await` its results. Throws an error if no match can be located.
     */
    async getHarness(query) {
        return __classPrivateFieldGet(this, _AngularContext_loader, "f").getHarness(query);
    }
    /**
     * Gets all component harnesses that match the query, wrapped for use in a fakeAsync test so that you do not need to `await` its results.
     */
    async getAllHarnesses(query) {
        return __classPrivateFieldGet(this, _AngularContext_loader, "f").getAllHarnesses(query);
    }
    /**
     * Advance time and trigger change detection. It is common to call this with no arguments to trigger change detection without advancing time.
     *
     * @param unit The unit of time `amount` represents. Accepts anything described in `@s-libs/s-core`'s [TimeUnit]{@linkcode https://simontonsoftware.github.io/s-js-utils/typedoc/enums/timeunit.html} enum.
     */
    tick(amount = 0, unit = 'ms') {
        if (isUndefined(window.Zone.current.get('FakeAsyncTestZoneSpec'))) {
            throw new Error('.tick() only works inside the .run() callback (because it needs to be in a fakeAsync zone)');
        }
        // To simulate real life, trigger change detection before processing macro tasks. To further simulate real life, wait until the micro task queue is empty.
        flushMicrotasks();
        this.runChangeDetection();
        tick(convertTime(amount, unit, 'ms'));
        this.runChangeDetection();
    }
    /**
     * This is a hook for subclasses to override. It is called during `run()`, before the `test()` callback. This implementation does nothing, but if you override this it is still recommended to call `super.init()` in case this implementation does something in the future.
     */
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    init() { }
    runChangeDetection() {
        this.inject(ApplicationRef).tick();
    }
    /**
     * Runs post-test verifications. This base implementation runs [HttpTestingController.verify]{@linkcode https://angular.io/api/common/http/testing/HttpTestingController#verify} and {@linkcode MockErrorHandler.verify}. Unlike {@linkcode #cleanUp}, it is OK for this method to throw an error to indicate a violation.
     */
    verifyPostTestConditions() {
        this.inject(HttpTestingController).verify();
        this.inject(MockErrorHandler).verify();
    }
    /**
     * Performs any cleanup needed at the end of each test. This base implementation calls {@linkcode https://angular.io/api/core/testing/discardPeriodicTasks|discardPeriodicTasks} and [flush]{https://angular.io/api/core/testing/flush|flush} to avoid an error from the `fakeAsync` zone.
     */
    cleanUp() {
        discardPeriodicTasks();
        flush();
        __classPrivateFieldSet(AngularContext, _a, undefined, "f", _AngularContext_current);
    }
}
_a = AngularContext, _AngularContext_loader = new WeakMap(), _AngularContext_instances = new WeakSet(), _AngularContext_runWithMockedTime = function _AngularContext_runWithMockedTime(test) {
    // https://github.com/angular/angular/issues/31677#issuecomment-573139551
    const { now } = performance;
    spyOn(performance, 'now').and.callFake(() => Date.now());
    jasmine.clock().install();
    fakeAsync(() => {
        jasmine.clock().mockDate(this.startTime);
        test();
    })();
    jasmine.clock().uninstall();
    performance.now = now;
};
_AngularContext_current = { value: void 0 };

/**
 * A mock method call that was made and is ready to be answered. This interface allows access to the underlying <code>jasmine.CallInfo</code>, and allows resolving or rejecting the asynchronous call's result.
 */
class TestCall {
    constructor(deferred, autoTick) {
        this.deferred = deferred;
        this.autoTick = autoTick;
    }
    /**
     * Resolve the call with the given value.
     */
    flush(value) {
        this.deferred.resolve(value);
        this.maybeTick();
    }
    /**
     * Reject the call with the given reason.
     */
    error(reason) {
        this.deferred.reject(reason);
        this.maybeTick();
    }
    maybeTick() {
        if (this.autoTick) {
            AngularContext.getCurrent()?.tick();
        }
    }
}

var _AsyncMethodController_instances, _AsyncMethodController_spy, _AsyncMethodController_testCalls, _AsyncMethodController_ensureCallInfoIsSet, _AsyncMethodController_makeArgumentMatcher, _AsyncMethodController_stringifyUserInput;
/**
 * Controller to be used in tests, that allows for mocking and flushing any asynchronous function. For example, to mock the browser's paste functionality:
 *
 * ```ts
 *  it('can paste', () => {
 *   const clipboard = navigator.clipboard;
 *   const ctx = new AngularContext();
 *
 *   // mock the browser API for pasting
 *   const controller = new AsyncMethodController(clipboard, 'readText');
 *   ctx.run(() => {
 *     // BEGIN production code that copies to the clipboard
 *     let pastedText: string;
 *     clipboard.readText().then((text) => {
 *       pastedText = text;
 *     });
 *     // END production code that copies to the clipboard
 *
 *     controller.expectOne([]).flush('mock clipboard contents');
 *
 *     // BEGIN expect the correct results after a successful copy
 *     expect(pastedText!).toBe('mock clipboard contents');
 *     // END expect the correct results after a successful copy
 *   });
 * });
 * ```
 */
class AsyncMethodController {
    /**
     * If you are using an `AngularContext`, the default behavior is to automatically call `.tick()` after each `.flush()` and `.error()` to trigger promise handlers and changed detection. This is the normal production behavior of asynchronous browser APIs. However, if zone.js does not patch the function you are stubbing, change detection would not run automatically. In that case you many want to turn off this behavior by passing the option `autoTick: false`. See the list of functions that zone.js patches [here](https://github.com/angular/angular/blob/master/packages/zone.js/STANDARD-APIS.md).
     */
    constructor(obj, methodName, { autoTick = true } = {}) {
        _AsyncMethodController_instances.add(this);
        _AsyncMethodController_spy.set(this, void 0);
        _AsyncMethodController_testCalls.set(this, []);
        // Note: it wasn't immediately clear how avoid `any` in this constructor, and this will be invisible to users. So I gave up. (For now?)
        __classPrivateFieldSet(this, _AsyncMethodController_spy, spyOn(obj, methodName), "f");
        __classPrivateFieldGet(this, _AsyncMethodController_spy, "f").and.callFake((async () => {
            const deferred = new Deferred();
            __classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f").push(new TestCall(deferred, autoTick));
            return deferred.promise;
        }));
    }
    /**
     * Expect that a single call was made that matches the given parameters or predicate, and return its mock. If no such call was made, or more than one, fail with a message including `description`, if provided.
     */
    expectOne(match, description) {
        const matches = this.match(match);
        if (matches.length !== 1) {
            throw new Error(buildErrorMessage({
                matchType: 'one matching',
                itemType: 'call',
                matches,
                stringifiedUserInput: __classPrivateFieldGet(this, _AsyncMethodController_instances, "m", _AsyncMethodController_stringifyUserInput).call(this, match, description),
            }));
        }
        return matches[0];
    }
    /**
     * Expect that no calls were made which match the given parameters or predicate. If a matching call was made, fail with a message including `description`, if provided.
     */
    expectNone(match, description) {
        const matches = this.match(match);
        if (matches.length > 0) {
            throw new Error(buildErrorMessage({
                matchType: 'zero matching',
                itemType: 'call',
                stringifiedUserInput: __classPrivateFieldGet(this, _AsyncMethodController_instances, "m", _AsyncMethodController_stringifyUserInput).call(this, match, description),
                matches,
            }));
        }
    }
    /**
     * Search for calls that match the given parameters or predicate, without any expectations.
     */
    match(match) {
        __classPrivateFieldGet(this, _AsyncMethodController_instances, "m", _AsyncMethodController_ensureCallInfoIsSet).call(this);
        let filterFn;
        if (Array.isArray(match)) {
            filterFn = __classPrivateFieldGet(this, _AsyncMethodController_instances, "m", _AsyncMethodController_makeArgumentMatcher).call(this, match);
        }
        else {
            filterFn = match;
        }
        return remove(__classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f"), (testCall) => filterFn(testCall.callInfo));
    }
    /**
     * Verify that no unmatched calls are outstanding. If any are, fail with a message indicating which calls were not matched.
     */
    verify() {
        if (__classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f").length) {
            __classPrivateFieldGet(this, _AsyncMethodController_instances, "m", _AsyncMethodController_ensureCallInfoIsSet).call(this);
            let message = buildErrorMessage({
                matchType: 'no open',
                itemType: 'call',
                stringifiedUserInput: undefined,
                matches: __classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f"),
            }) + ':';
            for (const testCall of __classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f")) {
                message += `\n  ${stringifyArgs(testCall.callInfo.args)}`;
            }
            throw new Error(message);
        }
    }
}
_AsyncMethodController_spy = new WeakMap(), _AsyncMethodController_testCalls = new WeakMap(), _AsyncMethodController_instances = new WeakSet(), _AsyncMethodController_ensureCallInfoIsSet = function _AsyncMethodController_ensureCallInfoIsSet() {
    for (let i = 1; i <= __classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f").length; ++i) {
        const testCall = nth(__classPrivateFieldGet(this, _AsyncMethodController_testCalls, "f"), -i);
        testCall.callInfo ?? (testCall.callInfo = nth(__classPrivateFieldGet(this, _AsyncMethodController_spy, "f").calls.all(), -i));
    }
}, _AsyncMethodController_makeArgumentMatcher = function _AsyncMethodController_makeArgumentMatcher(args) {
    return (callInfo) => isEqual(callInfo.args, args);
}, _AsyncMethodController_stringifyUserInput = function _AsyncMethodController_stringifyUserInput(match, description) {
    if (isUndefined(description)) {
        if (Array.isArray(match)) {
            description = 'Match by arguments: ' + stringifyArgs(match);
        }
        else {
            description = 'Match by function: ' + match.name;
        }
    }
    return description;
};
function stringifyArgs(args) {
    return JSON.stringify(args);
}

// adapted from https://github.com/ngneat/spectator/blob/e13c9554778bdb179dfc7235aedb4b3b90302850/projects/spectator/src/lib/mock.ts
/**
 * Creates a new object with jasmine spies for each method in `type`.
 *
 * ```ts
 * class Greeter {
 *   greet(name: string) {
 *     return `Hello, ${name}!`;
 *   }
 * }
 *
 * const spyObject = createSpyObject(Greeter);
 * spyObject.greet.and.returnValue("Hello, stub!");
 * expect(spyObject.greet("Eric")).toBe("Hello, stub!");
 * expectSingleCallAndReset(spyObject.greet, "Eric");
 * ```
 */
function createSpyObject(type) {
    const mock = {};
    for (let proto = type.prototype; proto !== null; proto = Object.getPrototypeOf(proto)) {
        for (const key of functions(proto)) {
            mock[key] = jasmine.createSpy(key);
        }
    }
    return mock;
}

/**
 * Expects the spy to have been called the exact number of times as the argument lists provided, with those arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1)
 * spy(2, 3)
 * spy()
 * expectCallsAndReset(spy, [1], [2, 3], []);
 * ```
 */
function expectCallsAndReset(spy, ...allArgs) {
    expect(spy).toHaveBeenCalledTimes(allArgs.length);
    expect(spy.calls.allArgs()).toEqual(allArgs);
    spy.calls.reset();
}

/**
 * Expects exactly one call to have been made to a jasmine spy, for it to have received the given arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1, 2);
 * expectSingleCallAndReset(spy, 1, 2); // pass
 * expectSingleCallAndReset(spy, 1, 2); // fail
 *
 * spy(3);
 * spy(4);
 * expectSingleCallAndReset(spy, 3); // fail
 * ```
 */
function expectSingleCallAndReset(spy, ...params) {
    expect(spy).toHaveBeenCalledTimes(1);
    expect(spy).toHaveBeenCalledWith(...params);
    spy.calls.reset();
}

// eslint-disable-next-line @angular-eslint/prefer-on-push-component-change-detection -- change detection is carefully orchestrated in the typescript */
class WrapperComponent {
    constructor() {
        this.inputs = {};
        this.styles = {};
    }
    static wrap(componentType, unboundInputs) {
        const selector = getSelector(componentType);
        const inputMetas = getInputMetas(componentType).filter(({ property }) => !unboundInputs.includes(property));
        TestBed.overrideComponent(WrapperComponent, {
            set: { template: buildTemplate(selector, inputMetas) },
        });
        return inputMetas.map((meta) => meta.property);
    }
}
WrapperComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.0.3", ngImport: i0, type: WrapperComponent, deps: [], target: i0.ɵɵFactoryTarget.Component });
WrapperComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.0.3", type: WrapperComponent, selector: "ng-component", ngImport: i0, template: '', isInline: true });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.0.3", ngImport: i0, type: WrapperComponent, decorators: [{
            type: Component,
            args: [{ template: '' }]
        }] });
function getSelector(componentType) {
    const annotations = Reflect.getOwnPropertyDescriptor(componentType, '__annotations__');
    assert(annotations, 'That does not appear to be a component');
    let selector = annotations.value.find((decorator) => decorator.selector)?.selector;
    if (!isValidSelector(selector)) {
        selector = 's-libs-component-under-test';
        TestBed.overrideComponent(componentType, { set: { selector } });
    }
    return selector;
}
function isValidSelector(selector) {
    if (!selector) {
        return false;
    }
    try {
        document.createElement(selector);
        return true;
    }
    catch {
        return false;
    }
}
function getInputMetas(componentType) {
    let metas;
    const superType = Object.getPrototypeOf(componentType.prototype)?.constructor;
    if (isDefined(superType)) {
        metas = getInputMetas(superType);
    }
    else {
        metas = [];
    }
    // I tried making this support inputs with special characters in their names, but it turns out that *Angular* can only support that when using AOT. So our *dynamic* wrapper cannot.
    forOwn(componentType.propDecorators, (decorators, property) => {
        const inputDecorators = decorators.filter((decorator) => decorator.type.prototype.ngMetadataName === 'Input');
        for (const decorator of inputDecorators) {
            const binding = decorator.args?.[0] ?? property;
            metas.push({ property, binding });
        }
    });
    return metas;
}
function buildTemplate(selector, inputMetas) {
    const bindingStrings = inputMetas.map(({ binding, property }) => `[${binding}]="inputs.${String(property)}"`);
    return `
    <div class="s-libs-dynamic-wrapper" [ngStyle]="styles">
      <${selector} ${bindingStrings.join(' ')}></${selector}>
    </div>
  `;
}

/**
 * Provides the foundation for an opinionated pattern for component tests.
 *
 * - Includes all features from {@link AngularContext}
 * - Automatically creates your component at the beginning of `run()`.
 * - Sets up Angular to call `ngOnChanges()` like it would in production. This is not the case if you use the standard `TestBed.createComponent()` directly.
 * - Wraps your component in a parent that you can easily style however you like.
 * - Lets you use {@link https://material.angular.io/cdk/test-harnesses/overview|component harnesses} in the `fakeAsync` zone, which is normally a challenge.
 * - Automatically disables animations.
 * - Causes async {@link https://angular.io/api/core/APP_INITIALIZER APP_INITIALIZER}s to complete before instantiating the component. Two caveats:
 *   - this requires all work in your initializers to complete with a call to `tick()`
 *   - this requires delaying app initialization until inside the `fakeAsync` zone, i.e. with the callback to {@link #run}. If you have async initializers, you must be careful not to do things that finalize the app setup before then, such as {@link #inject}.
 *
 * A very simple example:
 * ```ts
 * @Component({ template: 'Hello, {{name}}!' })
 * class GreeterComponent {
 *   @Input() name!: string;
 * }
 *
 * it('greets you by name', () => {
 *   const ctx = new ComponentContext(GreeterComponent);
 *   ctx.assignInputs({ name: 'World' });
 *   ctx.run(() => {
 *     expect(ctx.fixture.nativeElement.textContent).toBe('Hello, World!');
 *   });
 * });
 * ```
 *
 * A full example, with routing and a component harness. This is the full code for a tiny Angular app:
 * ```ts
 * /////////////////
 * // app-context.ts
 *
 * // To re-use your context setup, make a subclass of ComponentContext to
 * // import into any spec
 * class AppContext extends ComponentContext<AppComponent> {
 *   constructor() {
 *     super(AppComponent, {
 *       imports: [
 *         // This is your production `AppModule`. Make sure it exports `AppComponent`
 *         AppModule,
 *         // Import `routes` from your `app-routing.module.ts`
 *         RouterTestingModule.withRoutes(routes),
 *       ],
 *     });
 *   }
 * }
 *
 * ////////////////////////
 * // app.component.spec.ts
 *
 * describe('AppComponent', () => {
 *   let ctx: AppContext;
 *   beforeEach(() => {
 *     ctx = new AppContext();
 *   });
 *
 *   it('can navigate to the first page', () => {
 *     ctx.run(() => {
 *       ctx.getHarness(AppComponentHarness).navigateToFirstPage();
 *       expect(ctx.fixture.nativeElement.textContent).toContain(
 *         'First works!',
 *       );
 *     });
 *   });
 * });
 *
 * ///////////////////////////
 * // app.component.harness.ts
 *
 * // A simple component harness to demonstrate its integration with component contexts. Note that everything here is asynchronous, but is treated as synchronous in tests. This is a trick in AngularContexts so harnesses can be used in the fakeAsync zone.
 * class AppComponentHarness extends ComponentHarness {
 *   static hostSelector = 'app-root';
 *
 *   private getFirstPageLink = this.locatorFor('a');
 *
 *   async navigateToFirstPage(): Promise<void> {
 *     const link = await this.getFirstPageLink();
 *     await link.click();
 *   }
 * }
 *
 * /////////////////////
 * // first.component.ts
 *
 * // A minimal component for demonstration purposes
 * @Component({ template: '<p>First works!</p>' })
 * class FirstComponent {}
 *
 * ///////////////////
 * // app.component.ts
 *
 * // A minimal app component with routing for demonstration purposes
 * @Component({
 *   selector: 'app-root',
 *   template: `
 *     <a routerLink="/first-page">First Page</a>
 *     <router-outlet></router-outlet>
 *   `,
 * })
 * class AppComponent {}
 *
 * ////////////////////////
 * // app-routing.module.ts
 *
 * // exported for use in tests
 * const routes: Routes = [{ path: 'first-page', component: FirstComponent }];
 *
 * @NgModule({
 *   imports: [RouterModule.forRoot(routes)],
 *   exports: [RouterModule],
 * })
 * class AppRoutingModule {}
 *
 * ////////////////
 * // app.module.ts
 *
 * // A minimal app module. Notice the added export.
 * @NgModule({
 *   declarations: [AppComponent, FirstComponent],
 *   imports: [BrowserModule, BrowserAnimationsModule, AppRoutingModule],
 *   bootstrap: [AppComponent],
 *   exports: [AppComponent], // exported for use in tests
 * })
 * class AppModule {}
 * ```
 */
class ComponentContext extends AngularContext {
    /**
     * @param componentType `run()` will create a component of this type before running the rest of your test.
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link NoopAnimationsModule}, in addition to those provided by {@link AngularContext}.
     * @param unboundInputs By default a synthetic parent component will be created that binds to all your component's inputs. Pass input names here that should NOT be bound. This is useful e.g. to test the default value of an input.
     */
    constructor(componentType, moduleMetadata = {}, unboundInputs = []) {
        // TODO: once cleanup of contexts is not so touchy, move this below super() and use shortcut `private` declarations on constructor params
        const inputProperties = WrapperComponent.wrap(componentType, unboundInputs);
        super(extendMetadata(moduleMetadata, {
            imports: [NoopAnimationsModule],
            declarations: [WrapperComponent, componentType],
        }));
        this.componentType = componentType;
        this.inputProperties = new Set(inputProperties);
        this.inputs = {};
        this.wrapperStyles = {};
    }
    /**
     * Assign css styles to the div wrapping your component. Can be called before or during `run()`. Accepts an object with the same structure as the {@link https://angular.io/api/common/NgStyle|ngStyle directive}.
     *
     * ```ts
     * ctx.assignWrapperStyles({
     *   width: '400px',
     *   height: '600px',
     *   margin: '20px auto',
     *   border: '1px solid',
     * });
     * ```
     */
    assignWrapperStyles(styles) {
        Object.assign(this.wrapperStyles, styles);
        if (this.isInitialized()) {
            this.flushStylesToWrapper();
            this.tick();
        }
    }
    /**
     * Assign inputs passed into your component. Can be called before `run()` to set the initial inputs, or within `run()` to update them and trigger all the appropriate change detection and lifecycle hooks.
     */
    assignInputs(inputs) {
        for (const key of keys(inputs)) {
            if (!this.inputProperties.has(key)) {
                throw new Error(`Cannot bind to "${String(key)}" (it is not an input, or you passed it in \`unboundProperties\`)`);
            }
        }
        Object.assign(this.inputs, inputs);
        if (this.isInitialized()) {
            this.flushInputsToWrapper();
            this.tick();
        }
    }
    /**
     * Use within `run()` to get your instantiated component that is on the page.
     */
    getComponentInstance() {
        return this.fixture.debugElement.query(By.directive(this.componentType))
            .componentInstance;
    }
    /**
     * Constructs and initializes your component. Called during `run()` before it executes the rest of your test. Runs in the same `fakeAsync` zone as the rest of your test.
     */
    init() {
        super.init();
        // Inject something (I think anything) to finalize the compiler and trigger any registered APP_INITIALIZER. Then `tick()` to (hopefully) run it, assuming the user is only doing things that will be flushed right away.
        this.inject(ApplicationInitStatus);
        tick(); // can't use `this.tick()` until the component exists
        this.fixture = TestBed.createComponent(WrapperComponent);
        this.flushStylesToWrapper();
        this.flushInputsToWrapper();
        this.fixture.detectChanges();
        this.tick();
    }
    runChangeDetection() {
        this.fixture.detectChanges();
    }
    /**
     * Performs any cleanup needed at the end of each test. This implementation destroys {@link fixture} and calls the super implementation.
     */
    cleanUp() {
        this.fixture.destroy();
        super.cleanUp();
    }
    isInitialized() {
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions -- this actually can be undefined, but typing doesn't reflect it because once everything is initialized its defined
        return !!this.fixture;
    }
    flushInputsToWrapper() {
        Object.assign(this.getWrapperComponentInstance().inputs, this.inputs);
    }
    flushStylesToWrapper() {
        Object.assign(this.getWrapperComponentInstance().styles, this.wrapperStyles);
    }
    getWrapperComponentInstance() {
        return this.fixture.componentInstance;
    }
}

/**
 * A class very similar to Angular's [TestRequest]{@linkcode https://angular.io/api/common/http/testing/TestRequest} for use with an {@linkcode AngularContext}. If you are using an `AngularContext`, this will trigger change detection automatically after you flush a response, like production behavior.
 *
 * Though it is possible to construct yourself, normally an instance of this class is obtained from {@linkcode expectRequest()}.
 *
 * ```ts
 * const ctx = new AngularContext();
 * ctx.run(() => {
 *   ctx
 *     .inject(HttpClient)
 *     .get('http://example.com', { params: { key: 'value' } })
 *     .subscribe();
 *   const request = expectRequest<string>('GET', 'http://example.com', {
 *     params: { key: 'value' },
 *   });
 *   request.flush('my response body');
 * });
 * ```
 */
class SlTestRequest {
    constructor(req) {
        this.req = req;
        /**
         * The underlying [TestRequest]{@linkcode https://angular.io/api/common/http/testing/TestRequest} object from Angular.
         */
        this.request = this.req.request;
    }
    /**
     * Resolve the request with the given body and options, like [TestRequest.flush()]{@linkcode https://angular.io/api/common/http/testing/TestRequest#flush}.
     */
    flush(body, opts) {
        this.req.flush(body, opts);
        this.tickIfPossible();
    }
    /**
     * Convenience method to flush an error response.
     */
    flushError(status = 500, { statusText = 'simulated test error', body = null, } = {}) {
        this.req.flush(body, { status, statusText });
        this.tickIfPossible();
    }
    /**
     * Returns whether the request has been cancelled.
     */
    isCancelled() {
        return this.req.cancelled;
    }
    tickIfPossible() {
        AngularContext.getCurrent()?.tick();
    }
}

let matchCount;
let pendingRequests;
/**
 * This convenience method is similar to [HttpTestingController.expectOne()]{@linkcode https://angular.io/api/common/http/testing/HttpTestingController}, with extra features. The returned request object will automatically trigger change detection when you flush a response, just like in production.
 *
 * This method is opinionated in that you must specify all aspects of the request to match. E.g. if the request specifies headers, you must also specify them in the arguments to this method.
 *
 * This method only works when you are using an {@linkcode AngularContext}.
 *
 * ```ts
 * const ctx = new AngularContext();
 * ctx.run(() => {
 *   ctx
 *     .inject(HttpClient)
 *     .get('http://example.com', { params: { key: 'value' } })
 *     .subscribe();
 *   const request = expectRequest<string>('GET', 'http://example.com', {
 *     params: { key: 'value' },
 *   });
 *   request.flush('my response body');
 * });
 * ```
 */
function expectRequest(method, url, options = {}) {
    expect().nothing(); // convince jasmine we are expecting something
    const ctx = AngularContext.getCurrent();
    assert(ctx, 'expectRequest only works while an AngularContext is in use');
    const opts = { method, url, params: {}, headers: {}, body: null, ...options };
    try {
        return matchRequest(ctx, opts);
    }
    catch (error) {
        console.error('Expected 1 request to match:', opts, 'Actual pending requests:', pendingRequests);
        throw new Error(`Expected 1 matching request, found ${matchCount}. See details logged to the console.`);
    }
}
function matchRequest(ctx, options) {
    const controller = ctx.inject(HttpTestingController);
    matchCount = 0;
    pendingRequests = [];
    return new SlTestRequest(controller.expectOne((req) => {
        pendingRequests.push(req);
        const found = isMatch(req, options);
        if (found) {
            ++matchCount;
        }
        return found;
    }));
}
function isMatch(req, options) {
    return (req.method === options.method &&
        req.url === options.url &&
        matchAngularHttpMap(req.params, options.params) &&
        matchAngularHttpMap(req.headers, options.headers) &&
        isEqual(req.body, options.body));
}
function matchAngularHttpMap(actual, expected) {
    const actualObj = mapAsKeys(actual.keys(), (key) => actual.get(key));
    return isEqual(actualObj, expected);
}

/**
 * Provides some shorthand utilities that component harnesses may want.
 */
class ComponentHarnessSuperclass extends ContentContainerComponentHarness {
    /**
     * Searches for an instance of the component corresponding to the given harness type under the document root element, and returns a {@link ComponentHarness} for that instance. If multiple matching components are found, a harness for the first one is returned. If no matching component is found, an error is thrown.
     */
    async getTopLevelHarness(predicate) {
        const loader = await this.getTopLevelLoader();
        return loader.getHarness(predicate);
    }
    /**
     * Searches for all instances of the component corresponding to the given harness type under the document root element, and returns a list {@link ComponentHarness} for each instance.
     */
    async getAllTopLevelHarnesses(predicate) {
        const loader = await this.getTopLevelLoader();
        return loader.getAllHarnesses(predicate);
    }
    /**
     * Gets a {@link HarnessLoader} for the document root element. This loader can be used for elements that a component creates outside its own root element (e.g. by appending to document.body).
     */
    async getTopLevelLoader() {
        return this.documentRootLocatorFactory().rootHarnessLoader();
    }
}

/**
 * Designed for temporary use when a test using `fakeAsync` is giving you "Error: X timer(s) still in the queue", and you don't know what they are. Call this at the top of the test to see everything that adds a timer printed to the JS console. In Chrome you can click the function in the console output to be taken directly to its definition in the Sources tab.
 *
 * Be sure to call this _inside_ the test, not before, to get the most useful output.
 *
 * @returns a function to stop the logging
 *
 * ```ts
 * it('definitely does not set any timeouts', fakeAsync(() => {
 *   const stopLogging = logTimers();
 *
 *   doTheRestOfMyTest();
 *
 *   stopLogging();
 * }));
 * ```
 */
function logTimers() {
    const restoreTimeouts = wrapMethod(window, 'setTimeout', {
        before(...args) {
            console.log('setTimeout(', ...args, ')');
        },
    });
    const restoreIntervals = wrapMethod(window, 'setInterval', {
        before(...args) {
            console.log('setInterval(', ...args, ')');
        },
    });
    return () => {
        restoreTimeouts();
        restoreIntervals();
    };
}

/**
 * A helper to take some boilerplate out of RxJS's {@link https://rxjs-dev.firebaseapp.com/guide/testing/marble-testing marble testing} API.
 *
 * ```ts
 * it("doubles numbers", marbleTest(({ cold, expectObservable }) => {
 *   const source = cold("-1-2-3-|");
 *   const expected = "   -2-4-6-|");
 *   expectObservable(source.pipe(double())).toBe(expected);
 * }));
 * ```
 */
function marbleTest(callback) {
    const testScheduler = new TestScheduler((actual, expected) => {
        expect(actual).toEqual(expected);
    });
    return () => testScheduler.run((helpers) => callback({ ...helpers, testScheduler }));
}

/**
 * Use this when you want to write test code that doesn't actually run, instead relying only on your static tools like Typescript or a linter to raise errors.
 *
 * ```ts
 * function reject<T>(array: T[], predicate: (value: T) => boolean): T[] {
 *   return array.filter((value) => !predicate(value));
 * }
 *
 * it('requires the predicate type to match the array type', () => {
 *   staticTest(() => {
 *     // @ts-expect-error -- mismatch of number array w/ string function
 *     reject([1, 2, 3], (value: string) => value === '2');
 *   });
 * }); * ```
 */
function staticTest(_test) {
    expect().nothing();
}

/*
 * Public API Surface of ng-dev
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AngularContext, AsyncMethodController, ComponentContext, ComponentHarnessSuperclass, SlTestRequest, TestCall, createSpyObject, expectCallsAndReset, expectRequest, expectSingleCallAndReset, logTimers, marbleTest, staticTest };
//# sourceMappingURL=s-libs-ng-dev.mjs.map
